/*
Copyright (c) Nmap project and its contributors
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions 
are met:

 1. Redistributions of source code must retain the above copyright 
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.
 3. Neither the name of Nmap nor the names of its contributors 
    may be used to endorse or promote products derived from this 
    software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <libpq-fe.h>

#include <request.h>
#include <config.h>
#include <scan.h>
#include <debug.h>
#include <logger.h>
#include <db.h>

extern PGconn*		dbconn;
extern config_t*	config;

/*
 * Create scans needed by periodic scans
 * For example, if a periodic scan need a scan every 3 second, 
 * this function is used to create each of them.
 * The argument is the father scanid, used to create the inherited
 * scans. All father options are parsed and we create
 * an inherited scans with the same options.
 */
int create_inherited( const int father )
{
	char		buf[1024];
	int			uidowner;
	int			scanid;
	char		target[1024];
	char		*tmp;
	char		*optval;
	char		*optname;
	time_t		cepoch;
	int 		i;

	PGresult   *res;
	PGresult   *res2;

	if( ! getdb( dbconn ) )
	{
		DEBUGMSG( "getdb failed\n");
	}

	/*
	 * We will take the current epoch as reference
	 */
	if( time( &cepoch ) == -1 )
	{
		return( -1 );
	}

	/*
	 * Take the target and user-id from the periodic scan
	 */
	memset( buf , '\0' , 1024 );
	sprintf( buf , "DECLARE create_inherited_1 CURSOR FOR select target , uidowner FROM scans s WHERE id=%d" , father );
	DEBUGMSG( buf );
	res = PQexec( dbconn , buf );
	if( PQresultStatus( res ) != PGRES_COMMAND_OK )
	{
		PQclear( res );
		return( -2 );
	}

	res = PQexec( dbconn , "FETCH ALL in create_inherited_1");
	if( PQresultStatus( res ) != PGRES_TUPLES_OK )
	{
		PQclear( res );
		return( -3 );
	}
	
	if( PQntuples( res ) != 1 )
	{
		PQclear( res );
		return( -4 );
	}
	else
	{
		memset( target , '\0' , 1024 );
		tmp = PQgetvalue( res , 0 , 0 );
		strncpy( target , tmp , 1023 );
		tmp = PQgetvalue( res , 0 , 1 );
		uidowner = atoi( tmp );
	}
	PQclear( res );



	/*
	 * Now, we create an inherited scan with a temporary state
	 * If we don't use temporary state, the cgi-app will see it before options
	 * are added. So, it's better to put an temporary state that is not recognized
	 * by the web-app. We will update it when options are completed
	 */
	memset( buf , '\0' , 1024 );
	sprintf( buf , 
			"INSERT INTO scans( target , type , state , create_date , father , uidowner) VALUES( '%s' , %d , %d , NOW() , %d , %d )",
			target ,
			TYPE_INHERITED,
			STATE_TEMP,
			father ,
			uidowner );
	DEBUGMSG( buf );

	res = PQexec( dbconn , buf );
	if( PQresultStatus( res ) != PGRES_COMMAND_OK )
	{
		PQclear( res );
		return( -5 );
	}
	PQclear( res );


	/*
	 * We created a new inherited scan, but we don't have his id.
	 * So, we take the scan-id of this new scan.
	 * We need it, because in the scan_opts table, we have to declare the scanid
	 */
	memset( buf , '\0' , 1024 );
	sprintf( buf , "DECLARE create_inherited_2 CURSOR FOR select id FROM scans s WHERE father=%d ORDER BY id DESC LIMIT 1" , father );
	DEBUGMSG( buf );
	res = PQexec( dbconn , buf );
	if( PQresultStatus( res ) != PGRES_COMMAND_OK )
	{
		PQclear( res );
		return( -6 );
	}

	res = PQexec( dbconn , "FETCH ALL in create_inherited_2");
	if( PQresultStatus( res ) != PGRES_TUPLES_OK )
	{
		PQclear( res );
		return( -7 );
	}
	if( PQntuples( res ) != 1 )
	{
		PQclear( res );
		return( -8 );
	}
	else
	{
		tmp = PQgetvalue( res , 0 , 0 );
		scanid = atoi( tmp );
	}
	PQclear( res );


	/*
	 * Recopy each options of the father and
	 * add it to the inherited scan
	 */
	memset( buf , '\0' , 1024 );
	sprintf( buf , "DECLARE create_inherited_3 CURSOR FOR select optname , value FROM scan_opts WHERE sid=%d" , father );
	DEBUGMSG( buf );
	res = PQexec( dbconn , buf );
	if( PQresultStatus( res ) != PGRES_COMMAND_OK )
	{
		PQclear( res );
		return( -9 );
	}

	res = PQexec( dbconn , "FETCH ALL in create_inherited_3");
	if( PQresultStatus( res ) != PGRES_TUPLES_OK )
	{
		PQclear( res );
		return( -10 );
	}
	if( PQntuples( res ) <= 0 )
	{
		PQclear( res );
		return( -11 );
	}
	else
	{
		for( i = 0 ; i < PQntuples( res ) ; i++ )
		{
			optname = PQgetvalue( res , i , 0 );
			optval = PQgetvalue( res , i , 1 );
			memset( buf , '\0' , 1024 );
			sprintf( buf , "INSERT INTO scan_opts( sid , optname , value ) VALUES( %d , '%s' , '%s' )" , scanid , optname , optval );
			DEBUGMSG( buf );
			res2 = PQexec( dbconn , buf );
			if( PQresultStatus( res2 ) != PGRES_COMMAND_OK )
			{
				PQclear( res2 );
				return( -12 );
			}
			PQclear( res2 );
		}
		memset( buf , '\0' , 1024 );
		sprintf( buf , "INSERT INTO scan_opts( sid , optname , value ) VALUES( %d , '%s' , %li )" , scanid , DB_EPOCHOPT_STR , (unsigned long)cepoch );
		DEBUGMSG( buf );
		res2 = PQexec( dbconn , buf );
		if( PQresultStatus( res2 ) != PGRES_COMMAND_OK )
		{
			PQclear( res2 );
			return( -13 );
		}
		PQclear( res2 );
	}

	PQclear( res );


	/*
	 * Finally, change the state of the new inherited scan
	 * and tell that it's new.
	 */
	memset( buf , '\0' , 1024 );
	sprintf( buf , "UPDATE scans SET state=%d WHERE id=%d" , STATE_NEW , scanid );
	DEBUGMSG( buf );
	res = PQexec( dbconn , buf );
	if( PQresultStatus( res ) != PGRES_COMMAND_OK )
	{
		PQclear( res );
		return( -14 );
	}
	PQclear( res );


	if( ! releasedb( dbconn ) )
	{
		DEBUGMSG( "releasedb failed\n" );
	}

	return( 0 );
}


/*
 * This function search if a given periodic scan needs to perform
 * one inherited scan
 * First argument : scanid of the periodic scan
 * Second argument : the number of second which separate each inherited scan
 */
int need_new_scan( const int scanid , const int interval )
{
	int ret;
	char buf[1024];
	int ntuples;
	char *val;
	unsigned long	last_epoch;
	time_t			current_epoch;
	PGresult   *res;

	ret = 0;

	if( time( &current_epoch ) == -1 )
	{
		return( -1 );
	}

	if( ! getdb( dbconn ) )
	{
		return( -1 );
	}


	memset( buf , '\0' , 1024 );
	sprintf( buf , "DECLARE periodicnews CURSOR FOR select so.value FROM scans s , scan_opts so WHERE father=%d AND type=%d AND s.id = so.sid AND so.optname='%s' ORDER BY s.id DESC LIMIT 1" , scanid , TYPE_INHERITED , DB_EPOCHOPT_STR );
	DEBUGMSG( buf );
	res = PQexec( dbconn , buf );
	if( PQresultStatus( res ) != PGRES_COMMAND_OK )
	{
		PQclear( res );
		return( -3 );
	}
	PQclear( res );

	res = PQexec( dbconn , "FETCH ALL in periodicnews");
	if( PQresultStatus( res ) != PGRES_TUPLES_OK )
	{
		PQclear( res );
		return( -4 );
	}
	
	ntuples = PQntuples( res );
	if( ntuples == 0 )
		ret = 1;
	else
	{
		val = PQgetvalue( res , 0 , 0 );
		last_epoch = atoll( val );
		if( current_epoch > ( last_epoch + interval ) )
			ret = 1;
	}
	PQclear( res );

	if( ! releasedb( dbconn ) )
	{
		return( -1 );
	}

	return( ret );
}





/*
 * Create new inherited scan
 * Browse all periodic scan and create inherited if needed
 */
void new_inherited()
{
	int i;
	int interval;
	int scanid;
	char buf[1024];
	char *val;
	int t;
	PGresult   *res;

	if( ! getdb( dbconn ) )
	{
		DEBUGMSG( "getdb error" );
	}

	memset( buf , '\0' , 1024 );
	sprintf( buf , "DECLARE periodics CURSOR FOR select s.id , so.value FROM scans s , scan_opts so WHERE type=%d AND s.id = so.sid AND so.optname='%s'" , TYPE_PERIODIC , DB_PERIODICOPT_STR );
	DEBUGMSG( buf );
	res = PQexec( dbconn , buf );
	if( PQresultStatus( res ) != PGRES_COMMAND_OK )
	{
		PQclear( res );
		return;
	}
	PQclear( res );

	res = PQexec( dbconn , "FETCH ALL in periodics");
	if( PQresultStatus( res ) != PGRES_TUPLES_OK )
	{
		PQclear( res );
		return;
	}

	if( ! releasedb( dbconn) )
	{
		DEBUGMSG( "Cannot release DB" );
	}

	if( PQntuples( res ) > 0 )
	{
		for (i = 0; i < PQntuples( res ); i++ )
		{
			val = PQgetvalue( res , i , 0 );
			scanid = atoi( val );
			val = PQgetvalue( res , i , 1 );
			interval = atoi( val );

			t = need_new_scan( scanid , interval );
			if( t == 1 )
			{
				t = create_inherited( scanid );
				if( t != 0 )
					nmaplogcrit( "Cannot create inherited" );
				else
					nmaplog( "New inherited scan" );
			}
			else
			{
				if( t != 0 )
					nmaplogcrit( "Error when search new inherited scans" );
			}
		}
	}
	PQclear( res );

	DEBUGMSG( "END PERIODIC" );
}
